package element;

import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import parser.eparser.ElementParseException;

/**
 * This class represents an expression used to obtain an integer bean property.
 * As of 21/11/2006, some DeductiveEngine subclasses can be considered JavaBeans with some properties that can be obtained in this way
 * (e.g. RegionalErrorCorrectingDeductiveEngine with the property errorCount).
 * This class is designed so that the same mechanism can be applied to obtain properties from grammars, agendas or other participants
 * in the deductive parsing process if necessary.
 * 
 * @author Carlos Gmez
 * Created at mortadelo, 21/11/2006 17:44:02
 *
 */
public class GetIntegerBeanPropertyExpression extends ExpressionElement
{
	
	private String object;
	private String property;
	
	private static String[] beanNames = {"engine","cfg"};
	private static String[] beanAccessCodes = {"engine.EngineManager.getEngine()","grammar.CFGUtils.getGrammar()"};
	
    private GetIntegerBeanPropertyExpression( String object , String property )
    {
    	super ( new Element[0] ); //this expression has no children
    	this.object = object;
    	this.property = property;
    }
    
    /**
     * La expresin longitud de la cadena de entrada devuelve como cdigo
     * Java una llamada al cdigo que obtiene la propiedad que queremos.
     */
    public String getExpressionCode ( String[] memberNames )
    {
    	/*
        StringBuffer codeBuf = new StringBuffer();
        codeBuf.append("(IntElement.groundIntElement( ((Integer)");
        codeBuf.append(beanAccessCodes[findBeanName(object)]);
        codeBuf.append(".getClass().getMethod(\"");
        codeBuf.append("get");
        codeBuf.append(Character.toUpperCase(property.charAt(0))); //{this index is in range at this point}
        codeBuf.append(property.substring(1));
        codeBuf.append("\" , new Class[0]");
        codeBuf.append(")");
        codeBuf.append(".invoke( ");
        codeBuf.append(beanAccessCodes[findBeanName(object)]);
        codeBuf.append(", new Object[0]");
        codeBuf.append(")");
        codeBuf.append(").intValue()))");
        */
    	//doesn't work, code may throw exception and we don't catch it.
        
    	/*
    	StringBuffer codeBuf = new StringBuffer();
        codeBuf.append(beanAccessCodes[findBeanName(object)]);
        codeBuf.append(".");
        codeBuf.append("get");
        codeBuf.append(Character.toUpperCase(property.charAt(0))); //{this index is in range at this point}
        codeBuf.append(property.substring(1));
        codeBuf.append("()");
        codeBuf.append(")");
        */ //doesn't work, uncasted
    	
    	StringBuffer codeBuf = new StringBuffer();
    	codeBuf.append("((Integer)");
        codeBuf.append(beanAccessCodes[findBeanName(object)]);
        codeBuf.append(".");
        codeBuf.append("getProperty(\"");
        codeBuf.append(property);
        codeBuf.append("\")");
        codeBuf.append(")");

        return new String(codeBuf);
    	
    	
    }
    
    /**
     * Devuelve false.
     */
    public boolean isGround ()
    {
        return false;
    }
    
    /**
     * La clase resultante es la clase posicin de cadena de entrada.
     */
    public Class getResultingClass()
    {
        return element.IntElement.class;
    }
    
    private static int findBeanName ( String name )
    {
    	for ( int i = 0 ; i < beanNames.length ; i++ )
    	{
    		if ( beanNames[i].equals(name) ) return i;
    	}
    	return -1;
    }
    
    /**
     * El parsing es trivial, no nos fijamos en la cadena de entrada, simplemente
     * aceptamos aqullo que el usuario haya definido en el eparser como que representa
     * esta expresin.
     * Note that the getExpressionCode() method includes a more deep parsing, since
     * we need to determine 
     * @param s Cadena con la que se construir una expresin de este tipo.
     * @return Instancia de GetIntegerBeanPropertyExpression.
     */
    public static GetIntegerBeanPropertyExpression fromString ( String s ) throws ElementParseException
    {
    	GetIntegerBeanPropertyExpression expr;
    	StringTokenizer st = new StringTokenizer ( s , "." );
    	String object;
    	String property;
    	try
    	{
    	    object = st.nextToken();
    		property = st.nextToken();
    		expr = new GetIntegerBeanPropertyExpression(object,property);
    	}
    	catch ( NoSuchElementException nsee )
    	{
    		throw new ElementParseException("GetIntegerBeanPropertyExpression should be of the form object.Property");
    	}
    	if ( st.hasMoreTokens() || object.length() < 1 || property.length() < 1 )
    		throw new ElementParseException("GetIntegerBeanPropertyExpression should be of the form object.Property");
    	
    	if ( findBeanName(object) < 0 )
    		throw new ElementParseException("GetIntegerBeanPropertyExpression: bean \"" + object + "\" is unknown");
    	return expr;
    }
	

}
