
package element;

import java.util.*;

import parser.eparser.ElementParser;

/**
 * Este elemento representa una regla de gramtica independiente del contexto, a la cual "envuelve".
 * Como en todos los CompositeElements, esto se utiliza en el generador de cdigo;
 * pero ser sustituido por un array en el cdigo generado.
 * Se hace de esta manera porque a nivel de rbol de esquema tenemos que tener bien identificada
 * la clase de cada cosa para trabajar a la hora de generar el cdigo. Sin embargo, a nivel
 * de cdigo generado, la opcin ms eficiente es que los tipos de cada objeto compuesto
 * vayan "implcitos" y los elementos compuestos se manejen directamente como arrays de objetos,
 * que pueden ser a su vez elementos o arrays.
 * @author carlos
 *
 */
public class RuleWrapper extends CompositeElement 
{

	private Element[] syms;
	
	/**
	 * Construye un RuleWrapper a partir de sus elementos hijos.
	 * @param syms Elementos hijos de la regla.
	 */
	public RuleWrapper ( Element[] syms )
	{
		this.syms = syms;
	}
	
	/**
	 * Devuelve un array con los elementos hijos.
	 */
	public Element[] getRule()
	{
		return syms;
	}
	
	/**
	 * Devuelve siempre false.
	 */
	public boolean isGround()
	{
		return false;
	}
	
	/**
	 * Crea un RuleWrapper a partir de una cadena de la forma A -> B C D E.
	 * Ntese que es posible que tambin se creen instancias de smbolos.
	 * @param s Cadena para crear la regla.
	 * @return Regla creada.
	 */
	public static RuleWrapper fromString ( String s ) throws Exception
	{
		int flecha = s.indexOf("->");
		if ( flecha < 0 ) throw new Exception("Formato invlido en regla independiente del contexto"); /*invalid format*/
		String parteIzquierda = s.substring(0,flecha).trim();
		String parteDerecha = s.substring(flecha+2).trim();
		ArrayList simbolos = new ArrayList();
		StringTokenizer st1 = new StringTokenizer(parteIzquierda," ");
		StringTokenizer st2 = new StringTokenizer(parteDerecha," ");
		String symString;
		while ( st1.hasMoreTokens() )
		{
			symString = st1.nextToken();
			
			//Llamada al ElementParser para obtener el smbolo.
			Element sy = ElementParser.getCurrentInstance().parse(symString);
			
			simbolos.add(sy);
		}
		while ( st2.hasMoreTokens() )
		{
			symString = st2.nextToken();
			
			//Llamada al ElementParser para obtener el smbolo.
			Element sy = ElementParser.getCurrentInstance().parse(symString);
						
			simbolos.add(sy);
		}
		return new RuleWrapper((Element[])simbolos.toArray(new Element[0]));
	}
	
	/**
	 * Proporciona una representacin legible de la regla independiente del contexto
	 * envuelta en este wrapper.
	 */
	public String toString()
	{
		String s = syms[0].toString() + " ->";
		for ( int i = 1 ; i < syms.length ; i++ )
		{
			s += " ";
			s += syms[i].toString();
		}
		return s;
	}
	
	//from CompositeElement
	public int numChildren()
	{
		return syms.length;
	}
	
	//from CompositeElement
	public Element getChild ( int i )
	{
		return syms[i];
	}
	
}
