package engine;

import java.util.List;

import schema.step.DeductiveStep;
import schema.item.Items;
import schema.item.ItemHandler;
import schema.step.StepHandler;
import java.util.List;
import element.SpecialElement;
import element.IntElement;
import element.StringPosition;

import agenda.Agenda;


/**
 * This class overrides the process() method of the standard deductive engine to provide a version of the deductive engine with support
 * for regional error correction.
 * This machine works by increasing an error counter or decreasing a retract counter when the agenda is empty but no final (goal) items have
 * been found. However, while in RegionalErrorCorrectingDeductiveEngine these counters are stored in the deductive engine, in this case the
 * counters are issued as items of the form [ Counters , #e-1 , #e , #r , #d ] where the three integer elements in the item represent the error
 * count, the retract count and detection count, respectively (the latter two are string positions). In this way, the agenda doesn't have to be
 * reprocessed in every iteration of the error-correcting engine (i.e. each time counters change)
 * Of course, schemata to be executed in this machine should adequately handle the "Counters" items at their error-correcting deductive
 * steps.
 * 
 * @author Carlos Gmez
 * Created at mortadelo, 21/11/2006 13:02:13
 *
 */
public class RegionalErrorCorrectingDeductiveEngineWithCountersAsItems extends StandardDeductiveEngine
{
	

	public RegionalErrorCorrectingDeductiveEngineWithCountersAsItems ( Agenda agenda , List deductiveSteps , List goalConditions , ItemHandler itemHandler , StepHandler stepHandler )
	{
		super ( agenda , deductiveSteps , goalConditions , itemHandler , stepHandler );
	}
	
	
	/**
	 * Returns whether the given list contains final items according to the goal conditions associated to the schema
	 * this deductive engine was instantiated with.
	 */
	private boolean finalItemsFound( List lItems )
	{
		
		boolean found = false;
		for ( int i = 0 ; i < lItems.size() ; i++ )
		{
			Object[] item = (Object[]) lItems.get(i);
			if ( goalConditions == null || goalConditions.size() == 0 ) { found = true ; break; }
			else
			{
				for ( int j = 0 ; j < goalConditions.size() ; j++ )
				{
					List result = ((DeductiveStep)goalConditions.get(j)).applyTo(item);
					if ( result != null && result.size() > 0 )
					{
						found = true;
						break; //esto slo rompe el for externo
					}
				}
				if ( found ) break;
			}
		}
		return found;
		
	}
	
	
	private int retractCount = 0;
	private int errorCount = 0;
	private int rightmostPositionScanned = 0;
	private int detectionCount = 0;
	
	public int getRetractCount()
	{
		return retractCount;
	}
	
	public int getErrorCount()
	{
		return errorCount;
	}
	
	public int getDetectionCount()
	{
		return detectionCount;
	}
	
	
	/*
	public void setRightmostPositionScanned ( int pos )
	{
		this.rightmostPositionScanned = pos;
	}
	*/
	
	
	
	private int rightmostNextToExamine = 0;
	
	public int getRightmostPositionScanned ( )
	{
		List items = itemHandler.getAll();
		for ( ; rightmostNextToExamine < items.size() ; rightmostNextToExamine++ )
		{
			Object[] item = (Object[]) items.get(rightmostNextToExamine);
			if ( item[0] instanceof element.SpecialElement && ((element.SpecialElement)item[0]).toString().equals("Parse") )
			{
				if ( item[3] instanceof element.StringPosition )
				{
					int newVal = ((element.StringPosition)item[3]).getValue();
					if ( rightmostPositionScanned < newVal ) rightmostPositionScanned = newVal;
				}
			}
		}
		System.out.println("Rightmost pos " + rightmostPositionScanned);
		System.out.println("I analyzed " + items.size() + " items");
		return rightmostPositionScanned;
	}
	
	
	public List preprocess()
	{	
		return super.process();
	}
	
	
	public static int MAX_ERRORS = 2;
	
	
	public void issueCounterItem()
	{
		SpecialElement countersElement = SpecialElement.fromString("Counters");
		IntElement errCountElementMinus1 = IntElement.groundIntElement(errorCount-1);
		IntElement errCountElement = IntElement.groundIntElement(errorCount);
		StringPosition retCountElement = StringPosition.groundFromValue(retractCount);
		StringPosition detCountElement = StringPosition.groundFromValue(detectionCount);
		System.out.println("Issuing item: " + Items.toString(new Object[]{countersElement,errCountElementMinus1,errCountElement,retCountElement,detCountElement}));
		Object[] item = new Object[]{countersElement,errCountElementMinus1,errCountElement,retCountElement,detCountElement};
		agenda.add( item );
		itemHandler.addItem ( item );
		
	}
	
	
	public List process()
	{
	
		retractCount = 0;
		errorCount = 0;
		rightmostPositionScanned = 0;
		rightmostNextToExamine = 0;
		detectionCount = 0;
		
		boolean done = false;
		
		for (;;)
		{
			
			System.out.println("Going to launch standard processing. [e="+errorCount+", r="+retractCount+", d="+detectionCount+"]");
			System.out.println("Agenda size: " + agenda.size());
			
			//execute the standard deductive engine until agenda is empty
			List lItems = super.process();
			
			System.out.println("Standard processing done. [e="+errorCount+", r="+retractCount+", d="+detectionCount+"]");
			
			//check whether final items have been found
			if ( finalItemsFound(lItems) ) return lItems;
			else
			{
				
				System.out.println("Final items not found.");
				
				//Re-insert items into agenda for error parsing
				//not necessary in this subclass
				/*
				for ( int i = 0 ; i < lItems.size() ; i++ )
					agenda.add(lItems.get(i));
				*/
				
				if ( retractCount > 0 )
				{
					retractCount--;
					
					//[added 2006-11-19]
					//?? -> this way it can go right if necessary.
					int newDetection = getRightmostPositionScanned();
					if ( newDetection > detectionCount )
					{
						detectionCount = newDetection;
						retractCount = detectionCount;
					}
					
				
				}
				else
				{
					detectionCount = getRightmostPositionScanned();
					retractCount = detectionCount;
					errorCount++;
					
					
					//Re-insert items into agenda for error parsing
					//necessary in this subclass only when error count is incremented
					
					for ( int i = 0 ; i < lItems.size() ; i++ )
						agenda.add(lItems.get(i));
					
					
					
					if ( errorCount > MAX_ERRORS )
					{
						System.out.println("No parsers found for max error count " + MAX_ERRORS + ". Sorry.");
						return lItems;
					}
					
				}
				
				issueCounterItem();
				
				System.out.println("Retr " + retractCount + ", err " + errorCount);
				
			}
			
		}
		
		
	}
	
	public Object getProperty ( String key )
	{
		if ( key.equals("retractCount") )
			return new Integer(retractCount);
		if ( key.equals("errorCount") )
			return new Integer(errorCount);
		if ( key.equals("detectionCount") )
			return new Integer(detectionCount);
		return null;
	}
	

	
	
}