package element.tag;

import java.util.ArrayList;
import java.util.List;

import element.HashIndexable;
import element.SimpleElement;
import element.Symbol;
import element.SymbolTable;
import element.fs.PlainFeatureStructure;
import grammar.tag.TreeAdjoiningGrammar;
import grammar.tag.TreeAdjoiningGrammarWithFeatureStructures;

/**
 * This class represents a tree-adjoining grammar tree.
 * It contains the node to symbol mappings associated to the tree.
 * It does not contain the rules, which are represented by TAGTreeRules linked to this tree.
 * @author Carlos
 *
 */
public class TAGTree extends SimpleElement implements HashIndexable
{
	
	private String name;
	
	private List mapping = new ArrayList();
	
	private TAGTreeNode artificialRoot = new TAGTreeNode( this , 0 );
	
	private TreeAdjoiningGrammar tag;
	
	//private TAGTreeNode anchor; //may be null
	
	private List anchorNodes = new ArrayList();
	
	private boolean isInitial;
	
	private boolean ground = true;
	
	private String family;
	
	//As symbolForRoot() turns out to be a bottleneck, called millions of times,
	//and the root node never changes, we cache it.
	private Symbol cachedRootSymbol;
	
	public Object getHashKey()
	{
		return name;
	}
	
	public static TAGTree nonGroundFromString ( String s )
	{
		TAGTree t = new TAGTree ( null , s );
		t.ground = false;
		return t;
	}
	
	public TAGTree ( TreeAdjoiningGrammar tag , String name )
	{
		this.tag = tag;
		this.name = name;
	}
	
	public void setFamily( String family )
	{
		this.family = family;
	}
	
	public String getFamily()
	{
		return family;
	}
	
	public TreeAdjoiningGrammar getGrammar()
	{
		return tag;
	}
	
	public String getName()
	{
		return name;
	}
	
	public int numNodes()
	{
		return mapping.size()+1; //includes artificial root node
	}
	
	public TAGTreeNode getArtificialRoot()
	{
		return artificialRoot;
	}
	
	public Symbol symbolForRoot()
	{
		if ( cachedRootSymbol == null )
		{
			if ( numNodes() > 1 )	
				cachedRootSymbol = symbolForNode(1);
			else
				cachedRootSymbol = null;
		}
		return cachedRootSymbol;
	}
	
	public TAGTreeNode addNodeFor ( Symbol sym )
	{
		mapping.add(sym);
		TAGTreeNode ttn = new TAGTreeNode ( this , mapping.size() );
		return ttn;
	}
	
	public TAGTreeNode addNodeFor ( Symbol sym , boolean nullAdjunction )
	{
		mapping.add(sym);
		TAGTreeNode ttn = new TAGTreeNode ( this , mapping.size() , nullAdjunction );
		return ttn;
	}
	
	public TAGTreeNode addNodeFor ( Symbol sym , boolean nullAdjunction , boolean subsNode )
	{
		mapping.add(sym);
		TAGTreeNode ttn = new TAGTreeNode ( this , mapping.size() , nullAdjunction , subsNode );
		return ttn;
	}
	
	/*
	public void setAnchor ( TAGTreeNode anchor )
	{
		this.anchor = anchor;
	}
	
	public TAGTreeNode getAnchor()
	{
		return anchor;
	}
	*/
	
	public void addAnchor ( TAGTreeNode anchor )
	{
		anchorNodes.add(  anchor );
	}
	
	public List getAnchorNodes ( )
	{
		return anchorNodes;
	}
	
	public boolean hasNoAnchor ( )
	{
		return anchorNodes == null || anchorNodes.size() == 0;
	}
	
	public boolean hasAnchor ( Symbol s )
	{
		for ( int i = 0 ; i < anchorNodes.size() ; i++ )
		{
			TAGTreeNode anchorNode = (TAGTreeNode) anchorNodes.get(i);
			if ( anchorNode.getAssociatedSymbol() == s )
				return true;
		}
		return false;
		
	}
	
	public boolean hasValidAnchorFor ( Symbol s , PlainFeatureStructure pfs )
	{
		for ( int i = 0 ; i < anchorNodes.size() ; i++ )
		{
			TAGTreeNode anchorNode = (TAGTreeNode) anchorNodes.get(i);
			if ( anchorNode.getAssociatedSymbol() == s )
			{
				TreeAdjoiningGrammarWithFeatureStructures ftag = (TreeAdjoiningGrammarWithFeatureStructures) tag;
				PlainFeatureStructure top = ftag.getTopFeatureStructure ( anchorNode );
				PlainFeatureStructure bottom = ftag.getBottomFeatureStructure( anchorNode );
				PlainFeatureStructure unif = PlainFeatureStructure.unify ( PlainFeatureStructure.unify(top,bottom) , pfs );
				if ( unif != null ) return true;
			}
		}
		return false;
	}
	
	public int addMapping ( Symbol sym )
	{
		mapping.add(sym);
		return mapping.size(); //returns the number assigned to the symbol.
	}
	
	public Symbol symbolForNode ( int nNode )
	{
		if ( nNode == 0 )
			return SymbolTable.instance().symbolFor("Troot"); //implicit 0 -> T
		else
			return (Symbol) mapping.get(nNode-1);
	}

	public boolean isInitial() 
	{
		return isInitial;
	}
	
	public String toString()
	{
		return name;
	}

	public void setInitial (boolean isInitial) 
	{
		this.isInitial = isInitial;
	}

	public boolean isGround() 
	{
		return ground;
	}

}
