/*
 * Created on 09-ago-2004 11:22:41
 *
 */
package generator;

import element.CompositeElement;
import element.SequenceElement;
import element.SimpleElement;
import element.ExpressionElement;

/**
 * <p>Generador de cdigo para el constructor del paso deductivo.</p>
 * <p>Esta clase visita la condicin lateral del paso deductivo, y basndose en ella
 * genera el cdigo del constructor.</p>
 * <p>Created 09-ago-2004 11:22:41</p>
 * @author carlos
 */
public class ConstructorAssignVisitor extends CodeGeneratingVisitor 
{

    /**
     * Genera nmeros para aadir al nombre de las variables enteras que se utilizan
     * como ndices para los bucles de matching de los elementos Sequence. De este
     * modo, si matcheamos varios Sequence, podemos dar nombres distintos a las
     * variables.
     */
    private int tempVariableIndex = 0;
    
    /**
     * A esta cadena se le dar un valor cuando este visitante visite un elemento
     * de tipo Sequence. Este valor ser el nombre de la variable que contiene
     * el nmero de items que han emparejado con el Sequence. Tal nombre es necesario
     * almacenarlo para que leerlo cuando estemos visitando el CompositeElement que
     * contiene el Sequence, pues lo utilizaremos para aadirlo al ndice del array.
     */
    private String indexIncrement = null;
    
    /**
     * Devuelve un nombre para una variable temporal. Se garantiza que en cada
     * llamada se devuelve un nombre distinto.
     * @return El nombre para una variable temporal.
     */
    private String nextTempVariableName()
    {
    	return ( "aseq_temp_" + (tempVariableIndex++) );
    }	
	
	/**
	 * @param objectName Nombre del item argumento del constructor en el cdigo.
	 * @param toAppendTo Buffer en el que se almacenar el cdigo generado.
	 */
	public ConstructorAssignVisitor(String objectName, StringBuffer toAppendTo , DeductiveStepCodeGenerator cg) 
	{
		super(objectName, toAppendTo,cg);
	}

	public void visitSimpleElement(SimpleElement elt)
	{
		Attribute attr = generator.attributeFor ( elt );
		if ( attr == null ) /*no hay atributo => no asignamos nada (es un singleton, no aparece en antec o conseq)*/ ;
		else 
		{
			toAppendTo.append( attr.getName() + " = " + " (" + attr.getClassName() + ") " + objectName + ";\n" );
			generator.setGround(attr,true); //marcamos que ese atributo ya tiene un valor desde que se ejecuta el constructor
		}
	}
	
	/**
	 * En este caso no se hace nada para los elementos compuestos salvo pasar a los hijos,
	 * pues las asignaciones en principio sern siempre de objetos elementales.
	 */
    public void visitCompositeElement ( CompositeElement ce )
    {
		String oldObjectName = this.objectName;
		indexIncrement = null;
		String increments = "";
		for ( int i = 0 ; i < ce.numChildren() ; i++ )
		{	
			if ( indexIncrement != null )
			{
				/**
				 * Si entramos en este if, es que el ltimo hijo procesado era un
				 * Sequence (multimatcher). En este caso, no debemos incrementar
				 * el ndice en 1, pues no sabemos en tiempo de generacin de cdigo
				 * si emparejaremos con un elemento, con cero o con ms de uno.
				 * El nombre de la variable que contendr el nmero de elementos con
				 * que emparejamos lo guarda el mtodo visitSequenceElement de 
				 * este mismo visitante en la variable indexIncrement. As pues,
				 * tenemos que sumar ese nombre a nuestro ndice en el cdigo generado.
				 */
				increments+=("+"+indexIncrement);
				indexIncrement = null;
			}
			this.objectName = "((Object[])"+oldObjectName + ")" + "["+i+increments+"]";
			ce.getChild(i).accept(this);
		}
		this.objectName = oldObjectName; //dejamos el visitante como estaba.
    }
	
	public void visitExpressionElement(ExpressionElement ee)
	{
	    System.err.println("[ERROR] Expressions not permitted in side conditions!");
	}
	
	public void visitSequenceElement(SequenceElement zome)
	{
		
		//El Sequence tiene que ser ground (no hay elementos no-ground en la condicin lateral).
		
		Attribute attr = generator.attributeFor ( zome );
		if ( attr == null ) ; /*no hay atributo => no asignamos nada (es un singleton, no aparece en antec o conseq)*/
		
		else
		{
			
			//Tomar el ndice de objectName.
			//Aqu partimos la cadena objectName para separar en partes
			//el ndice del array, la parte anterior, la parte posterior y el nombre
			//del array. Por ejemplo, si tenemos
			//miItem[k][6][i+1] => hacemos:
			//beforeIndex = miItem[k][6][
			//index = i+1
			//afterIndex = ]
			//arrayName = miItem[k][6] (el arrayName se usa para ver la longitud).
			int lastOpenBracket = objectName.lastIndexOf("[");
			int closedBracket = objectName.indexOf("]",lastOpenBracket);
			String beforeIndex = objectName.substring(0,lastOpenBracket+1);
			String index = objectName.substring(lastOpenBracket+1,closedBracket);
        	String afterIndex = objectName.substring(closedBracket);
        	String arrayName = beforeIndex.substring(0,beforeIndex.length()-1) + afterIndex.substring(1);
        
        	//obtenemos el nombre para el ndice de bucle de este multimatcher.
        	String loopIndexName = nextTempVariableName();
        
        	//Generamos el cdigo de un bucle que ir recorriendo elementos y aadindolos
        	//a la lista del multimatcher mientras sean de la clase asociada a ste.
        	//Ntese que en este caso no necesitamos estructuras try-catch
        	//ni cdigo de recuperacin, porque el matching de un Sequence no puede
        	//fallar: en el caso de que nos encontremos elementos cuya clase no
        	//coincide con la esperada por el Sequence, simplemente ste habr
        	//emparejado con cero elementos; pero empareja correctamente.
        	toAppendTo.append ( attr.getName() + " = new ArrayList();\n" );
        	toAppendTo.append ( "int " + loopIndexName + " = " + index + ";\n");
        	toAppendTo.append ( "while ( " + loopIndexName + " < " + arrayName + ".length && " + beforeIndex + loopIndexName + afterIndex + " instanceof " + zome.getAssociatedClass().getName() + " )\n" );
        	toAppendTo.append ( "{\n" );
        	toAppendTo.append ( attr.getName() + ".add(" + beforeIndex + loopIndexName + afterIndex + ");\n" );
        	toAppendTo.append ( loopIndexName + "++;\n");
        	toAppendTo.append ( "}\n");
        
        	//Seteamos indexIncrement, utilizado por el mtodo visitCompositeElement
        	//que llama a ste.
        	indexIncrement = loopIndexName;
        
        	//Indicamos en el generador de cdigo que la lista ya es ground.
        	generator.setGround ( attr , true );
		
		} //else if non singleton
	} //method visitSequenceElement
	
}
