/*
 * Created on 09-ago-2004
 *
 */
package generator;

import element.CompositeElement;
import element.ExpressionElement;
import element.SimpleElement;
import element.SequenceElement;

/**
 * <p>Esta clase genera el cdigo de chequeo de la precondicin del constructor, para
 * el mtodo constructorPrecondition.</p>
 * <p>Para ello, visita la condicin lateral, y genera el cdigo de chequear un item contra
 * la condicin lateral para ver si con ese item podemos construir el paso deductivo.</p>
 * 
 * <p>Created 09-ago-2004 11:04:30</p>
 * @author carlos
 */
public class ConstructorMatchVisitor extends CodeGeneratingVisitor 
{
	
	/**
	 * Genera nmeros para aadir al nombre de las variables enteras que se utilizan
	 * como ndices para los bucles de matching de los elementos Sequence. De este
	 * modo, si matcheamos varios Sequence, podemos dar nombres distintos a las
	 * variables.
	 */
	private int tempVariableIndex = 0;
	
	/**
	 * A esta cadena se le dar un valor cuando este visitante visite un elemento
	 * de tipo Sequence. Este valor ser el nombre de la variable que contiene
	 * el nmero de items que han emparejado con el Sequence. Tal nombre es necesario
	 * almacenarlo para leerlo cuando estemos visitando el CompositeElement que
	 * contiene el Sequence, pues lo utilizaremos para aadirlo al ndice del array.
	 */
	private String indexIncrement = null;
	
	/**
	 * Devuelve un nombre para una variable temporal. Se garantiza que en cada
	 * llamada se devuelve un nombre distinto.
	 * @return El nombre para una variable temporal.
	 */
	private String nextTempVariableName()
	{
		return ( "mseq_temp_" + (tempVariableIndex++) );
	}
	
	/**
	 * @param objectName Nombre del objeto que analizamos en el cdigo.
	 * @param toAppendTo Buffer al que se agregar el cdigo generado.
	 */
	public ConstructorMatchVisitor(String objectName, StringBuffer toAppendTo , DeductiveStepCodeGenerator cg) 
	{
		super(objectName, toAppendTo,cg);
	}
	
	/**
	 * Genera el cdigo para un elemento simple, que consiste en comprobar que la clase
	 * del elemento que analizamos en el cdigo coincide con la clase esperada (que es
	 * la de nuestro elemento).
	 */
	public void visitSimpleElement(SimpleElement elt) 
	{		
		if ( elt.isGround() )
		{
			toAppendTo.append("if ( !" + objectName + ".equals(" + elt.getExpressionCode() + ")) return false;\n");
		}
		else
		
		{
			toAppendTo.append("if ( " + objectName + ".getClass() != " + elt.getClass().getName() + ".class" + " ) return false;\n" );
		}
	}
	
	/**
	 * Genera el cdigo para un elemento compuesto, que consiste en comprobar que la longitud
	 * del array de objetos que analizamos en el cdigo coincide con la longitud esperada
	 * (que es el nmero de hijos de nuestro elemento compuesto), y despus generar el cdigo
	 * para cada uno de los elementos del array y los hijos correspondientes (para ello
	 * el nombre de objeto de cdigo referenciado por el visitante se va cambiando a objeto[i]
	 * a medida que se visitan los hijos).
	 */
	public void visitCompositeElement ( CompositeElement ce )
	{
		
		//check desired length
		int desiredLength = ce.numChildren();
		for ( int i = 0 ; i < ce.numChildren() ; i++ )
		{
			if ( ce.getChild(i) instanceof SequenceElement )
				desiredLength--;
		}
		if ( desiredLength == ce.numChildren() )
			toAppendTo.append ( "if ( ((Object[])" + objectName + ").length != " + ce.numChildren() + " ) return false;\n" );
		else
			toAppendTo.append ( "if ( ((Object[])" + objectName + ").length < " + desiredLength + " ) return false;\n" );
		
		String oldObjectName = this.objectName;
		indexIncrement = null;
		String increments = "";
		for ( int i = 0 ; i < ce.numChildren() ; i++ )
		{	
			if ( indexIncrement != null )
			{
				/**
				 * Si entramos en este if, es que el ltimo hijo procesado era un
				 * Sequence (multimatcher). En este caso, no debemos incrementar
				 * el ndice en 1, pues no sabemos en tiempo de generacin de cdigo
				 * si emparejaremos con un elemento, con cero o con ms de uno.
				 * El nombre de la variable que contendr el nmero de elementos con
				 * que emparejamos lo guarda el mtodo visitSequenceElement de 
				 * este mismo visitante en la variable indexIncrement. As pues,
				 * tenemos que sumar ese nombre a nuestro ndice en el cdigo generado.
				 */
				increments+=("+"+indexIncrement);
				indexIncrement = null;
			}
			this.objectName =  "((Object[])"+oldObjectName + ")" + "["+i+increments+"]";
			ce.getChild(i).accept(this);
		}
		this.objectName = oldObjectName; //dejamos el visitante como estaba.
	}
	
	/**
	 * Damos error, pues las condiciones laterales no pueden contener expresiones.
	 */
	public void visitExpressionElement(ExpressionElement ee)
	{
		System.err.println("[ERROR] Expressions not permitted in side conditions!");
	}
	
	public void visitSequenceElement ( SequenceElement zome )
	{
		//Tomar el ndice de objectName.
		//Aqu partimos la cadena objectName para separar en partes
		//el ndice del array, la parte anterior, la parte posterior y el nombre
		//del array. Por ejemplo, si tenemos
		//miItem[k][6][i+1] => hacemos:
		//beforeIndex = miItem[k][6][
		//index = i+1
		//afterIndex = ]
		//arrayName = miItem[k][6] (el arrayName se usa para ver la longitud).
		int lastOpenBracket = objectName.lastIndexOf("[");
		int closedBracket = objectName.indexOf("]",lastOpenBracket);
		String beforeIndex = objectName.substring(0,lastOpenBracket+1);
		String index = objectName.substring(lastOpenBracket+1,closedBracket);
		String afterIndex = objectName.substring(closedBracket);
		String arrayName = beforeIndex.substring(0,beforeIndex.length()-1) + afterIndex.substring(1);
		
		//obtenemos el nombre para el ndice de bucle de este multimatcher.
		String loopIndexName = nextTempVariableName();
		
		//Generamos el cdigo de un bucle que ir recorriendo elementos y no haciendo
		//nada mientras sean de la clase asociada al SequenceElement.
		//Ntese que en este caso no necesitamos estructuras try-catch
		//ni cdigo de recuperacin, porque el matching de un Sequence no puede
		//fallar: en el caso de que nos encontremos elementos cuya clase no
		//coincide con la esperada por el Sequence, simplemente ste habr
		//emparejado con cero elementos; pero empareja correctamente.
		toAppendTo.append ( "int " + loopIndexName + " = " + index + ";\n");
		toAppendTo.append ( "while ( " + loopIndexName + " < " + arrayName + ".length && " + beforeIndex + loopIndexName + afterIndex + " instanceof " + zome.getAssociatedClass().getName() + " )\n" );
		toAppendTo.append ( "{\n" );
		toAppendTo.append ( ";\n" );
		toAppendTo.append ( loopIndexName + "++;\n");
		toAppendTo.append ( "}\n");
		
		//Seteamos indexIncrement, utilizado por el mtodo visitCompositeElement
		//que llama a ste.
		indexIncrement = loopIndexName;		
		
	}

	
}


