package grammar;

import element.BooleanElement;
import element.Element;
import element.Symbol;
import element.SymbolTable;

/**
 * This class provides utilities for context-free grammars.
 * @author Carlos
 *
 */
public class CFGUtils 
{

	private static ContextFreeGrammar grammar;
	
	public static ContextFreeGrammar getGrammar()
	{
		return grammar;
	}
	
	public static void registerGrammar ( ContextFreeGrammar g )
	{
		grammar = g;
	}
	
	private static boolean[][] leftCornerTable;
	private static boolean[] terminalTable;
	private static boolean leftCornerTableInitted = false;
	private static boolean terminalTableInitted = false;
	
	public static void initLeftCornerTable()
	{
		
		System.out.println("Computing Left-Corner Relationships...");
		long iniTime = System.currentTimeMillis();
		
		int nSymbols = SymbolTable.instance().getLastId()+1;
		leftCornerTable = new boolean[nSymbols][nSymbols];
		
		//reflexivity
		for ( int i = 0 ; i < nSymbols ; i++ )
			leftCornerTable[i][i] = true;
		
		//transitivity
		boolean changed = true;
		while ( changed )
		{
			changed = false;
			for ( int i = 0 ; i < grammar.size() ; i++ )
			{
				Element[] rule = grammar.getEntity(i);
	
				for ( int aind = 0 ; aind < nSymbols ; aind++ )
				{
					if ( leftCornerTable[aind][((Symbol)rule[0]).getID()] )
					{
						if ( rule.length > 1 ) //non-epsilon rule
						{
							if ( leftCornerTable[aind][((Symbol)rule[1]).getID()] == false )
								changed = true;
							leftCornerTable[aind][((Symbol)rule[1]).getID()] = true;					
						}
					}
				}
			}
		}
		
		long elTime = System.currentTimeMillis() - iniTime;
		System.out.println("LC Relationships computed in " + elTime + "ms.");
		
	}
	
	public static BooleanElement leftCorner ( Symbol s1 , Symbol s2 )
	{
		if ( grammar == null )
			System.err.println("[ERROR] Left-corner query performed, but grammar not registered at CFGUtils class. Make sure your grammar class calls the CFGUtils.register(grammar) static method.");
		if ( !s1.isGround() || !s2.isGround() )
		{
			System.err.println("[ERROR] Left-Corner query performed on nonground symbol(s)");
		}
		if ( !leftCornerTableInitted )
		{
			initLeftCornerTable();
			leftCornerTableInitted = true;
		}
		boolean b = leftCornerTable[s1.getID()][s2.getID()];
		return BooleanElement.groundFromValue(b);
	}
	
	public static int countLeftCornerSetFor ( Symbol s1 )
	{
		if ( grammar == null )
			System.err.println("[ERROR] Left-corner query performed, but grammar not registered at CFGUtils class. Make sure your grammar class calls the CFGUtils.register(grammar) static method.");
		if ( !s1.isGround() )
		{
			System.err.println("[ERROR] Left-Corner query performed on nonground symbol");
		}
		if ( !leftCornerTableInitted )
		{
			initLeftCornerTable();
			leftCornerTableInitted = true;
		}
		boolean [] bs = leftCornerTable[s1.getID()];
		int counter = 0;
		for ( int i = 0 ; i < bs.length ; i++ )
		{
			if ( bs[i] )
				counter++;
		}
		return counter;
	}
	
	/**
	 * Creates a data structure to differenciate terminal symbols from nonterminals.
	 *
	 */
	public static void initTerminalTable()
	{
		
		System.out.println("Computing Terminal Table...");
		long iniTime = System.currentTimeMillis();
		
		int nSymbols = SymbolTable.instance().getLastId()+1;
		terminalTable = new boolean[nSymbols];
		
		//by default everything is a terminal
		for ( int i = 0 ; i < nSymbols ; i++ )
			terminalTable[i] = true;
		
		for ( int i = 0 ; i < grammar.size() ; i++ )
		{
			//symbols at left-hand sides of rules are nonterminals
			Element[] rule = grammar.getEntity(i);
			terminalTable[ ((Symbol)rule[0]).getID() ] = false;
		}
		
		long elTime = System.currentTimeMillis() - iniTime;
		System.out.println("Terminal Table computed in " + elTime + "ms.");
		
	}
	
	public static BooleanElement isTerminal ( Symbol s )
	{
		if ( grammar == null )
			System.err.println("[ERROR] Is-terminal query performed, but grammar not registered at CFGUtils class. Make sure your grammar class calls the CFGUtils.register(grammar) static method.");
		if ( !s.isGround() )
		{
			System.err.println("[ERROR] Is-terminal query performed on nonground symbol");
		}
		if ( !terminalTableInitted )
		{
			initTerminalTable();
			terminalTableInitted = true;
		}
		boolean b = terminalTable[s.getID()];
		return BooleanElement.groundFromValue(b);
	}
	
	
	
	
	
	
	
}
