/*
 * Created on 02-ago-2004
 *
 */
package parser.eparser;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import element.*;
import java.lang.reflect.*;




/**
 * ElementParser tiene la responsabilidad de hacer el "parsing" de los elementos que aparecen en
 * los esquemas de anlisis sintctico: por ejemplo smbolos, nmeros, reglas, etc. Esta
 * responsabilidad es externa al SchemaParser porque la idea es que el formato de dichos elementos
 * sea redefinible por el usuario (que podra necesitar crear nuevos tipos de elementos) mientras
 * que el formato de los esquemas de anlisis sintctico es fijo.
 * De este modo, lo que se proporcionar al usuario es una manera sencilla de definir expresiones
 * regulares para cada uno de los tipos de elemento.
 * El formato del fichero de definicin ser simplemente as:
 * regex
 * constructor
 * regex
 * constructor...
 * Por ejemplo:
 * [a-cA-C]
 * new Symbol(str,true);
 * [0-9]*
 * new Integer(...);
 * @author carlos
 *
 * Formato de las definiciones de elementos:
 * element.Symbol:<init>:[a-zA-Z]
 * element.IntElt:fromString:[0-9]
 * Campos en cada lnea separados por dos puntos.
 * Primer campo: nombre de la clase.
 * Segundo campo: <init> para constructor o nombre de un mtodo esttico que tome como
 * parmetro un String para construir la instancia de la clase.
 * Tercer campo: regex.
 */
public class ElementParser 
{
	
	/**Expresiones regulares correspondientes a cada una de las clases de elemento.*/
	private ArrayList regexs = new ArrayList();
	
	/**Clases de elemento para cada expresin regular.*/
	private ArrayList classes = new ArrayList();
	
	/**Mtodos para construir los objetos.*/
	/**Son objetos de tipo CreationMethodAdapter, que contienen dentro un mtodo esttico
	 * o un constructor y pueden ser invocados sin preocuparse del tipo concreto de mtodo
	 * que envuelven.*/
	private ArrayList methods = new ArrayList();
	
	/**
	 * Referencia al objeto ElementParser cuyo mtodo parse(String) est siendo
	 * ejecutado actualmente, si es que lo hay.
	 */
	private static ElementParser currentParser = null;
	
	/**
	 * Devuelve una referencia al objeto ElementParser cuyo mtodo parse(String) est
	 * siendo ejecutado. Esta clase no est preparada para multithreading y el mtodo
	 * parse de un ElementParser no debera llamar al mtodo parse de otro ElementParser,
	 * as que como mximo puede haber una instancia que cumpla esta caracterstica.
	 * Este mtodo es til para que mtodos de parsing para elementos, que sern llamados
	 * por el ElementParser, puedan pedir que una subcadena de la cadena que tienen que
	 * parsear sea parseada como elemento por el mismo ElementParser.
	 * @return Referencia al objeto ElementParser cuyo mtodo parse(String) est siendo
	 * ejecutado. Si no se est ejecutando ningn mtodo parse(String) [no est ahora
	 * mismo en la pila de ejecucin] no se garantiza un valor de retorno concreto.
	 */
	public static ElementParser getCurrentInstance()
	{
		return currentParser;
	}
	
	/**
	 * Agrega a las expresiones regulares reconocidas por este parser de elementos la dada.
	 * @param line Lnea con formato clase:mtodo:regex, donde mtodo puede ser el nombre
	 * de un mtodo static que tome como parmetro un String o <init> para referirse a 
	 * un constructor, que tambin deber tomar como parmetro un string.
	 */
	public void addLine ( String line ) throws ElementSpecFormatException
	{
		
		StringTokenizer colonTokenizer = new StringTokenizer(line,":");
		if ( colonTokenizer.hasMoreTokens() ) //lnea vaca no la consideramos
		{
			String className,methodName,regex;
			try
			{
				className = colonTokenizer.nextToken();
				methodName = colonTokenizer.nextToken();
				regex = colonTokenizer.nextToken(""); /*rest of string with a leading ":"*/
				regex = regex.substring(1); /*remove delimiter (: character)*/				
			}
			catch ( NoSuchElementException nsee )
			{
				throw new ElementSpecFormatException ( "Wrong format for element format specification line [" + line + "]: " + nsee );
			}
			//className,methodName,regex not null here
			
			//obtener clase, o tirar ClassNotFoundException
			Class cl;
			try
			{
				cl = Class.forName(className);
			}
			catch ( ClassNotFoundException cnfe )
			{
				throw new ElementSpecFormatException ( "Error when parsing element format specification line [" + line + "]: couldn't find class " + className + ": " + cnfe );
			}
			//cl not null here
			
			Pattern pat;
			//compilar regex, o tirar PatternSyntaxException
			try
			{
				pat  = Pattern.compile(regex);
			}
			catch ( PatternSyntaxException pse )
			{
				throw new ElementSpecFormatException ( "Error when parsing element format specification line [" + line + "]: regular expression format is invalid: " + pse );
			}
			//pat not null here
			
			boolean constructor = false; //si vamos a usar el constructor
			String classLastName=className; 
			StringTokenizer dotst = new StringTokenizer(className,".");
			while ( dotst.hasMoreTokens() ) classLastName=dotst.nextToken();
			if ( methodName.equals("<init>") || methodName.equals(classLastName) )
				constructor = true;
			
			if ( constructor )
			{
				//comprobar que la clase tiene un constructor que tome un String
				//o tirar NoSuchMethodException
				Constructor cons;
				try
				{
					cons = cl.getConstructor(new Class[]{String.class});
				}
				catch ( NoSuchMethodException nsme )
				{
					throw new ElementSpecFormatException ( "Error when parsing element format specification line [" + line + "]: couldn't find constructor with one String parameter for class " + className + ": " + nsme );
				}
				//si llegamos aqu, todo ha ido bien
				//agregamos esta especificacin de elemento a la lista
				regexs.add(pat);
				classes.add(cl);
				methods.add ( new CreationMethodAdapter(cons) );
			}
			else
			{
				//comprobar que la clase tiene un mtodo esttico de nombre methodName
				//que tome un String, o tirar NoSuchMethodException
				Method met;
				try
				{
					met = cl.getMethod(methodName,new Class[]{String.class});
					int mods = met.getModifiers();
					if ( !Modifier.isStatic(mods) )
						throw new ElementSpecFormatException ( "Error when parsing element format specification line [" + line + "]: found method " + methodName + ", but it's not static" );
				}
				catch ( NoSuchMethodException nsme )
				{
					throw new ElementSpecFormatException ( "Error when parsing element format specification line [" + line + "]: couldn't find static method with one String parameter for class " + className + ": " + nsme );
				}
				//si llegamos aqu, todo ha ido bien
				//agregamos esta especificacin de elemento a la lista
				regexs.add(pat);
				classes.add(cl);
				methods.add( new CreationMethodAdapter(met) );
			}
			
		} //if has any token
		
	} //end method
	
	/**
	 * Agrega a las expresiones regulares reconocidas por este parser de elementos
	 * las que aparecen en el fichero dado.
	 * @param file Fichero de lneas con formato clase:mtodo:regex, donde mtodo puede ser el nombre
	 * de un mtodo static que tome como parmetro un String o <init> para referirse a 
	 * un constructor, que tambin deber tomar como parmetro un string.
	 */
	public void addFile ( File file ) throws ElementSpecFormatException ,
	FileNotFoundException,
	IOException
	{
		BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream ( file )));
		String linea = "";
		while ( linea != null )
		{
			linea = br.readLine();
			if ( linea != null )
				addLine(linea);
		}
	}
	
	public ElementParser ( )
	{
		;
	}
	
	public ElementParser ( File f ) throws ElementSpecFormatException,
	FileNotFoundException,
	IOException
	{
		this();
		this.addFile(f);
	}
	
		
	/**
	 * Interpreta un string devolviendo el elemento correspondiente para el rbol del
	 * parsing schema.
	 * @param s Cadena que representa a un elemento.
	 * @return Elemento construido a partir de esa cadena, teniendo en cuenta las definiciones
	 * regulares del fichero con el que se defini este parser.
	 */
	public Element parse ( String s ) throws ElementParseException
	{
		currentParser = this;

			//assert ( regexs.size() == patterns.size() == methods.size() )
			for ( int i = 0 ; i < regexs.size() ; i++ )
			{
				if ( ((Pattern)regexs.get(i)).matcher(s).matches() )
				{
					CreationMethodAdapter met = (CreationMethodAdapter)methods.get(i);
					//tenemos el constructor: lo usamos.
					Element elt = null;
					try
					{
						elt = (Element) met.invoke(new Object[]{s});
					}
					catch ( InvocationTargetException exc )
					{
						exc.printStackTrace();
						throw new ElementParseException("El mtodo invocado para el elemento tir una excepcin",exc);
					}
					catch ( Exception exc )
					{
						
						throw new ElementParseException(exc);
					}
					return elt;
				}
			}
			//si llegamos aqu, ninguna regex empareja
			throw new ElementParseException("La cadena " + s + " no se corresponde con ninguno de los formatos definidos para elementos.");
	}
	
}
